#include "ttwav.h"
#include <stdio.h>
#include <stdlib.h>

TtWav::TtWav() {
  this->channel_count = 0;
  this->frequency     = 0;
  this->bit_rate      = 0;
  this->data_size     = 0;
  this->memory_size   = 0;
  this->data          = NULL;
  this->debug         = false;
}

TtWav::~TtWav() {
  delete this->data;
}

void TtWav::make_blank() {
  this->channel_count = 2;
  this->frequency     = 44100;
  this->bit_rate      = 16;
  this->data_size     = 0;
  this->memory_size   = 1024;
  if ( this->data != NULL ) { delete this->data; }
  this->data          = new char[this->memory_size];
  memset( this->data, 0, this->memory_size );
}

int TtWav::get_channel_count() { return this->channel_count; }
int TtWav::get_frequency()     { return this->frequency;     }
int TtWav::get_bit_rate()      { return this->bit_rate;      }
int TtWav::get_data_size()     { return this->data_size;     }
int TtWav::get_memory_size()   { return this->memory_size;   }

void TtWav::set_debug( bool flag ) { this->debug = flag; }

bool TtWav::read_file( char *filepath ) {
  FILE *file = fopen( filepath, "rb" );
  if ( file == NULL ) { return false; }

  bool status;

  int riff_header;
  status = this->read_int( &riff_header, file );
  if ( !(status) ) { goto ERR; }
  if ( riff_header != TtWav::RIFF_HEADER ) { goto ERR; }

  int file_size;
  status = this->read_int( &file_size, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "file size = %d\n", file_size + 8 ); }

  int wave_header;
  status = this->read_int( &wave_header, file );
  if ( !(status) ) { goto ERR; }
  if ( wave_header != TtWav::WAVE_HEADER ) { goto ERR; }

  // fmt`NɂȂ܂œǂݔ΂
  int fmt_chunk;
  for (;;) {
    status = this->read_int( &fmt_chunk, file );
    if ( !(status) ) { goto ERR; }
    if ( fmt_chunk == TtWav::fmt_CHUNK ) { break; }

    int tmp_size;
    status = this->read_int( &tmp_size, file );
    if ( !(status) ) { goto ERR; }
    if ( fseek( file, tmp_size, SEEK_CUR ) != 0 ) { goto ERR; }
  }
  
  int fmt_byte_size;
  status = this->read_int( &fmt_byte_size, file );
  if ( !(status) ) { goto ERR; }
  // g̑Ή̂߂ɍ폜
  // if ( fmt_byte_size != TtWav::fmt_BYTE_SIZE ) { goto ERR; }
  if ( this->debug ) { printf( "fmt_byte_size = %d\n", fmt_byte_size ); }

  int fmt_id;
  status = this->read_2byte( &fmt_id, file );
  if ( !(status) ) { goto ERR; }
  if ( fmt_id != TtWav::fmt_ID ) { goto ERR; }
  if ( this->debug ) { printf( "fmt_id = %d\n", fmt_id ); }

  status = this->read_2byte( &this->channel_count, file );
  if ( !(status) ) { goto ERR; }
  if ( this->channel_count != 1 && this->channel_count != 2 ) { goto ERR; }
  if ( this->debug ) { printf( "channel_count = %d\n", this->channel_count ); }

  status = this->read_int( &this->frequency, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "frequency = %d\n", this->frequency ); }

  int byte_per_sec;
  status = this->read_int( &byte_per_sec, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "byte_per_sec = %d\n", byte_per_sec ); }

  int block_size;
  status = this->read_2byte( &block_size, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "block_size = %d\n", block_size ); }

  status = this->read_2byte( &this->bit_rate, file );
  if ( !(status) ) { goto ERR; }
  if ( this->bit_rate != 8 && this->bit_rate != 16 ) { goto ERR; }
  if ( this->debug ) { printf( "bit_rate = %d\n", this->bit_rate ); }
  
  if ( block_size != ( this->bit_rate / 8 ) * this->channel_count ) { goto ERR; }
  if ( byte_per_sec != this->frequency * block_size ) { goto ERR; }

  // g̑Ή
  if ( fmt_byte_size - TtWav::fmt_BYTE_SIZE > 0 ) { 
    int extend_size;
    status = this->read_2byte( &extend_size, file );
    if ( !(status) ) { goto ERR; }
    if ( extend_size > 0 ) {
      char *extend_data = new char[extend_size];
      fread( extend_data, 1, extend_size, file );
      delete extend_data;
    }
  }
  
  // data`NɂȂ܂œǂݔ΂
  int data_chunk;
  for (;;) {
    status = this->read_int( &data_chunk, file );
    if ( !(status) ) { goto ERR; }
    if ( data_chunk == TtWav::data_CHUNK ) { break; }

    int tmp_size;
    status = this->read_int( &tmp_size, file );
    if ( !(status) ) { goto ERR; }
    if ( fseek( file, tmp_size, SEEK_CUR ) != 0 ) { goto ERR; }
  }
  
  status = this->read_int( &this->data_size, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "data_size = %d\n", this->data_size ); }

  this->memory_size = (int)(
    (double)(this->data_size) *
    (double)( 2 / this->channel_count ) *
    ( (double)44100 / (double)this->frequency ) *
    (double)( 16 / this->bit_rate )
    );
  this->memory_size = this->memory_size > this->data_size ? this->memory_size : this->data_size;
  if ( this->debug ) { printf( "memory_size = %d\n", this->memory_size ); }
  this->data = new char[this->memory_size];
  int readed;
  readed = fread( this->data, 1, this->data_size, file );
  if ( this->debug ) { printf( "readed = %d\n", readed ); }
  if ( readed != this->data_size ) { goto ERR; }

  if ( this->debug ) { printf( "data head = %p\n", *((short *)this->data) ); }

  this->to_16bit_rate();
  if ( this->debug ) { printf( "16bit changed\n" ); }
  this->to_2ch();
  if ( this->debug ) { printf( "2ch changed\n" ); }
  this->to_44kHz();
  if ( this->debug ) { printf( "44kHz changed\n" ); }

  if ( this->debug ) { printf( "data head = %p\n", *((short *)this->data) ); }


  fclose( file );
  return true;

ERR:
  if ( this->debug ) { printf( "error\n" ); }
  if ( file != NULL ) { fclose( file ); }
  return false;
}

bool TtWav::read_int( int *data, FILE *file ) {
  int buf;
  int readed = fread( (char *)&buf, 1, sizeof( int ), file );
  if ( readed != sizeof( int ) ) { return false; }
  *(data) = buf;
  return true;
}

bool TtWav::read_2byte( int *data, FILE *file ) {
  short buf;
  int readed = fread( (char *)&buf, 1, sizeof( short ), file );
  if ( readed != sizeof( short ) ) { return false; }
  *(data) = buf;
  return true;
}

bool TtWav::write_file( char *filepath ) {
  FILE *file = fopen( filepath, "wb" );
  if ( file == NULL ) { return false; }

  if ( this->debug ) { printf( "-----write----------------------------\n" ); }
  
  bool status;

  status = this->write_int( TtWav::RIFF_HEADER, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "\"RIFF\" written\n" ); }

  status = this->write_int( 36 + this->data_size, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "file_size written = %d\n", 36 + this->data_size ); }

  status = this->write_int( TtWav::WAVE_HEADER, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "\"WAVE\" written\n" ); }

  status = this->write_int( TtWav::fmt_CHUNK, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "\"fmt \" written\n" ); }

  status = this->write_int( TtWav::fmt_BYTE_SIZE, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "fmt_BYTE_SIZE written\n" ); }

  status = this->write_2byte( TtWav::fmt_ID, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "fmt_ID written\n" ); }

  status = this->write_2byte( this->channel_count, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "channel_count written = %d\n", this->channel_count ); }

  status = this->write_int( this->frequency, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "frequency written = %d\n", this->frequency ); }

  status = this->write_int( this->frequency * ( this->bit_rate / 8 ) * ( this->channel_count ), file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "byte_per_sec written = %d\n", this->frequency * ( this->bit_rate / 8 ) * ( this->channel_count ) ); }

  status = this->write_2byte( ( this->bit_rate / 8 ) * ( this->channel_count ), file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "block_size written = %d\n", ( this->bit_rate / 8 ) * ( this->channel_count ) ); }

  status = this->write_2byte( this->bit_rate, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "bit_rate written = %d\n", this->bit_rate ); }

  status = this->write_int( TtWav::data_CHUNK, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "\"data\" written\n" ); }

  status = this->write_int( this->data_size, file );
  if ( !(status) ) { goto ERR; }
  if ( this->debug ) { printf( "memory size is %d\n", this->memory_size ); }
  if ( this->debug ) { printf( "data_size written = %d\n", this->data_size ); }

  int written;
  written = fwrite( this->data, 1, this->data_size, file );
  if ( written != this->data_size ) { goto ERR; }
  if ( this->debug ) { printf( "data body written %d\n", written ); }

  fclose( file );
  return true;

ERR:
  if ( this->debug ) { printf( "error\n" ); }
  if ( file != NULL ) { fclose( file ); }
  return false;
}

bool TtWav::write_int( int data, FILE *file ) {
  int written = fwrite( (char *)&data, 1, sizeof( int ), file );
  if ( written != sizeof( int ) ) { return false; }
  return true;
}

bool TtWav::write_2byte( int data, FILE *file ) {
  short buf = (short)data;
  int written = fwrite( (char *)&buf, 1, sizeof( short ), file );
  if ( written != sizeof( short ) ) { return false; }
  return true;
}



void TtWav::to_16bit_rate() {
  if ( this->bit_rate == 8 ) {
    this->memory_extend( this->data_size * 2 );
    for ( int i = this->data_size - 1; i >= 0; --i ) {
      ((short *)this->data)[i] = (short)((((this->data[i] & 0x000000FF) - 128)) * 256);
    }
    this->bit_rate = 16;
    this->data_size *= 2;
  }
}

void TtWav::to_2ch() {
  if ( this->channel_count == 1 ) {
    this->memory_extend( this->data_size * 2 );
    for ( int i = ( this->data_size / 2 ) - 1; i >= 0; --i ) {
      ((short *)this->data)[i * 2 + 1] = ((short *)this->data)[i];
      ((short *)this->data)[i * 2    ] = ((short *)this->data)[i];
    }
    this->channel_count = 2;
    data_size *= 2;
  }
}

void TtWav::to_44kHz() {
  // člKvA
  if ( this->frequency != 44100 ) {
    int new_data_size = (int)((double)data_size * ( (double)44100 / (double)this->frequency ));
    new_data_size -= new_data_size % 4;
    this->memory_extend( new_data_size );
    if ( this->frequency < 44100 ) {
      for ( int i = (new_data_size / 4) - 1; i >= 0; --i ) {
        ((int *)this->data)[i] = ((int *)this->data)[(int)((double)i / ((double)44100 / (double)this->frequency))];
      }
    }
    else {
      for ( int i = 0; i < new_data_size / 4; ++i ) {
        ((int *)this->data)[i] = ((int *)this->data)[(int)((double)i / ((double)44100 / (double)this->frequency))];
      }
    }
    this->data_size = new_data_size;
    this->frequency = 44100;
  }
}

void TtWav::memory_extend( int size ) {
  if ( size > this->memory_size ) {
    int new_size;
    for ( new_size = this->memory_size; new_size < size; new_size *= 2 ) {}
    char *new_data = new char[new_size];
    memcpy( new_data, this->data, this->memory_size );
    memset( new_data + this->memory_size, 0, new_size - this->memory_size );
    delete this->data;
    this->data = new_data;
    this->memory_size = new_size;
    if ( this->debug ) { printf( "memory extended %d\n", new_size ); }
  }
}


void TtWav::change_volume( int percentage ) {
  if ( percentage == 100 ) { return; }
  for ( int i = 0; i < ( this->data_size / 2 ); ++i ) {
    int _tmp = (int)((double)((short *)this->data)[i] * ((double)percentage / (double)100));
    ((short *)this->data)[i] =
      _tmp > TtWav::VOL_MAX ? TtWav::VOL_MAX
        : _tmp < TtWav::VOL_MIN ? TtWav::VOL_MIN
          : (short)_tmp;
  }
  if ( this->debug ) { printf( "changed %d%%\n", percentage ); }
}

void TtWav::mixin_at( int position, TtWav *other ) {
  this->memory_extend( position * 2 * 2 + other->data_size );
  if ( this->debug ) { printf( "mixin begin %d %d\n", this->memory_size, other->data_size ); }

  for ( int i = 0; i < ( other->data_size / 2 ); ++i ) {
    int _tmp = (int)(((short *)this->data)[2 * position + i]) + (int)(((short *)other->data)[i]);
    ((short *)this->data)[2 * position + i] =
      _tmp > TtWav::VOL_MAX ? TtWav::VOL_MAX
        : _tmp < TtWav::VOL_MIN ? TtWav::VOL_MIN
          : (short)_tmp;
  }
  if ( this->data_size < position * 2 * 2 + other->data_size ) {
    this->data_size = position * 2 * 2 + other->data_size;
  }
  if ( this->debug ) { printf( "mixin end\n" ); }
}
